#include "common.h"

namespace cmmn {

	double * smpls_rho;
	int n_smpls;



	//
	//								Quick Sort
	//
	//      S. Kawai (1988).  Methods of programming (in Japanese). Tokyo: Iwanami-shoten.
	//		S. Ishihata (1989). Algorithms and Data Structure (in Japanese). Tokyo:Iwanami-shoten
	//
	void qSort(double * d, int Lb, int Ub){
		if (Lb < Ub){
			int p = (Lb + Ub) / 2;
			double c = d[p];
			int i = Lb;
			int j = Ub;
			do {
				while (d[i] < c) i++;
				while (c < d[j]) j--;
				if (i <= j){
					double v = d[i];
					d[i] = d[j];
					d[j] = v;
					i++;
					j--;
				}
			} while (i <= j);
			qSort(d, Lb, j);
			qSort(d, i, Ub);
		}
		return;
	}


	double CalcCor(int * iX, int * iY, int n){
		double * X = new double[n];
		double * Y = new double[n];
		double sumX = 0.0, sumY = 0.0;
		for (int i = 0; i < n; i++){
			X[i] = iX[i];
			sumX += X[i];
			Y[i] = iY[i];
			sumY += Y[i];
		}
		double meanX = sumX / n, meanY = sumY / n;
		double sumXX = 0.0, sumYY = 0.0, sumXY = 0.0;
		for (int i = 0; i < n; i++){
			sumXX += sqr(X[i] - meanX);
			sumYY += sqr(Y[i] - meanY);
			sumXY += (X[i] - meanX) * (Y[i] - meanY);
		}
		double r = sumXY / Math::Sqrt(sumXX * sumYY);
		delete[] X;
		delete[] Y;
		return r;
	}

	double phi(double x){
		return Math::Exp(-0.5 * x * x) / Math::Sqrt(2.0 * Math::PI);
	}

	/**************************************************************************

	Kerridge and Cook(1976, Biometrika, 63, 401-403)'s Algorithm

	**************************************************************************/

	double normal_0_to_z(double z) {
		double p;
		z = Math::Abs(z);
		if (z > 6.8) {
			double zz = z * z;
			double d = zz + 3 - 1 / (0.22 * zz + 0.704);
			double m = 1 - 1 / d;
			p = 0.5 - Math::Exp(-0.5 * zz) * m / (z * Math::Sqrt(2 * Math::PI));
		}
		else {
			p = 0.0;
			if (z > 0.0) {
				double zz4 = 0.25 * z * z;
				double n2 = 2.0;                // initial value of 2n          ; n = 1
				double theta2 = 1.0;			// initial value of theta(2n-2)
				double theta1 = zz4;			// initial value of theta(2n-1)
				p = 1.0;						// sum for n = 0
				double p_prev = 0.0;			// initial value of sub-sum
				while (true) {
					theta2 = zz4 * (theta1 - theta2) / n2;
					n2 += 1.0;
					p += theta2 / n2;
					if (p == p_prev) break;
					theta1 = zz4 * (theta2 - theta1) / n2;
					n2 += 1.0;
					p_prev = p;
				}
				p = z * Math::Exp(-0.5 * zz4) * p / Math::Sqrt(2.0 * Math::PI);
			};
		}
		return p;
	}

	double Cum_NormalKC(double z) {
		if (z >= 0.0) {
			if (z > 0.0) {
				double v = 0.5 + normal_0_to_z(z);
				if (v > 1.0) {
					return 1.0;
				}
				else {
					return v;
				};
			}
			else {
				return 0.5;
			};
		}
		else {
			double v = 0.5 - normal_0_to_z(Math::Abs(z));
			if (v < 0.0){
				return 0.0;
			}
			else {
				return v;
			};
		}
	}

	/*
            	|           |
    	Pnp     |    Pcp    |    Ppp
            	|           |
	---------------------------------------
            	|           |
       	Pnc     |    Pcc    |    Ppc
            	|           |
	---------------------------------------
            	|           |
    	Pnn     |    Pcn    |    Ppn
            	|           |
	*/

	double CalcPpp(double vrho, double va, double vb){
		Funcfpp Ffpp(vrho, vb);
		TAdaptiveGL<Funcfpp> adapgl(Ffpp, 1.0e-10, 5);
		double v = adapgl.integral(-Math::Exp(-va), 0.0);
		return v;
	}

	double CalcPpc(double vrho, double va, double vb1, double vb2){
		Funcfpc Ffpc(vrho, va);
		TAdaptiveGL<Funcfpc> adapgl(Ffpc, 1.0e-10, 5);
		double v = adapgl.integral( vb1, vb2 );
		return v;
	}

	double CalcPpn(double vrho, double va, double vb){
		Funcfpn Ffpn(vrho, vb);
		TAdaptiveGL<Funcfpn> adapgl(Ffpn, 1.0e-10, 5);
		double v = adapgl.integral(-Math::Exp(-va), 0.0);
		return v;
	}

	double CalcPcp(double vrho, double va1, double va2, double vb){
		Funcfcp Ffcp(vrho, vb);
		TAdaptiveGL<Funcfcp> adapgl(Ffcp, 1.0e-10, 5);
		double v = adapgl.integral( va1, va2 );
		return v;
	}

	double CalcPcc(double vrho, double va1, double va2, double vb1, double vb2){
		Funcfcc Ffcc(vrho, vb1, vb2);
		TAdaptiveGL<Funcfcc> adapgl(Ffcc, 1.0e-10, 5);
		double v = adapgl.integral(va1, va2);
		return v;
	}

	double CalcPcn(double vrho, double va1, double va2, double vb){
		Funcfcn Ffcn(vrho, vb);
		TAdaptiveGL<Funcfcn> adapgl(Ffcn, 1.0e-10, 5);
		double v = adapgl.integral(va1, va2);
		return v;
	}

	double CalcPnp(double vrho, double va, double vb){
		Funcfnp Ffnp(vrho, vb);
		TAdaptiveGL<Funcfnp> adapgl(Ffnp, 1.0e-10, 5);
		double v = adapgl.integral(0.0, Math::Exp(va));
		return v;
	}

	double CalcPnc(double vrho, double va, double vb1, double vb2){
		Funcfnc Ffnc(vrho, va);
		TAdaptiveGL<Funcfnc> adapgl(Ffnc, 1.0e-10, 5);
		double v = adapgl.integral(vb1, vb2);
		return v;
	}

	double CalcPnn(double vrho, double va, double vb){
		Funcfnn Ffnn(vrho, vb);
		TAdaptiveGL<Funcfnn> adapgl(Ffnn, 1.0e-10, 5);
		double v = adapgl.integral(0.0, Math::Exp(va));
		return v;
	}
    

	void CalcPxy::calculate(){
		for (int i = 1; i <= Kx; i++)
			for (int j = 1; j <= Ky; j++){
				if (i <= 1){
					if (j <= 1){
						Pxy[i - 1][j - 1] = CalcPnn(rho, Cx[0], Cy[0]);
					}
					else if (j >= Ky){
						Pxy[i - 1][j - 1] = CalcPnp(rho, Cx[0], Cy[Ky - 2]);
					}
					else{
						Pxy[i - 1][j - 1] = CalcPnc(rho, Cx[0], Cy[j - 2], Cy[j - 1]);
					}
				}
				else if (i >= Kx){
					if (j <= 1){
						Pxy[i - 1][j - 1] = CalcPpn(rho, Cx[Kx - 2], Cy[0]);
					}
					else if (j >= Ky){
						Pxy[i - 1][j - 1] = CalcPpp(rho, Cx[Kx - 2], Cy[Ky - 2]);
					}
					else{
						Pxy[i - 1][j - 1] = CalcPpc(rho, Cx[Kx - 2], Cy[j - 2], Cy[j - 1]);
					}
				}
				else{
					if (j <= 1){
						Pxy[i - 1][j - 1] = CalcPcn(rho, Cx[i - 2], Cx[i - 1], Cy[0]);
					}
					else if (j >= Ky){
						Pxy[i - 1][j - 1] = CalcPcp(rho, Cx[i - 2], Cx[i - 1], Cy[Ky - 2]);
					}
					else{
						Pxy[i - 1][j - 1] = CalcPcc(rho, Cx[i - 2], Cx[i - 1], Cy[j - 2], Cy[j - 1]);
					}
				}
			}
	}


	void CalcPxyCx::calculate(){
	//	for (int i = 1; i <= Kx; i++)
		for (int i = Ck + 1; i <= Ck + 2; i++)
			for (int j = 1; j <= Ky; j++){
				if (i <= 1){
					if (j <= 1){
						Pxy[i - 1][j - 1] = CalcPnn(rho, Cx[0], Cy[0]);
					}
					else if (j >= Ky){
						Pxy[i - 1][j - 1] = CalcPnp(rho, Cx[0], Cy[Ky - 2]);
					}
					else{
						Pxy[i - 1][j - 1] = CalcPnc(rho, Cx[0], Cy[j - 2], Cy[j - 1]);
					}
				}
				else if (i >= Kx){
					if (j <= 1){
						Pxy[i - 1][j - 1] = CalcPpn(rho, Cx[Kx - 2], Cy[0]);
					}
					else if (j >= Ky){
						Pxy[i - 1][j - 1] = CalcPpp(rho, Cx[Kx - 2], Cy[Ky - 2]);
					}
					else{
						Pxy[i - 1][j - 1] = CalcPpc(rho, Cx[Kx - 2], Cy[j - 2], Cy[j - 1]);
					}
				}
				else{
					if (j <= 1){
						Pxy[i - 1][j - 1] = CalcPcn(rho, Cx[i - 2], Cx[i - 1], Cy[0]);
					}
					else if (j >= Ky){
						Pxy[i - 1][j - 1] = CalcPcp(rho, Cx[i - 2], Cx[i - 1], Cy[Ky - 2]);
					}
					else{
						Pxy[i - 1][j - 1] = CalcPcc(rho, Cx[i - 2], Cx[i - 1], Cy[j - 2], Cy[j - 1]);
					}
				}
			}
	}

	void CalcPxyCy::calculate(){
		for (int i = 1; i <= Kx; i++)
//			for (int j = 1; j <= Ky; j++){
			for (int j = Ch + 1; j <= Ch + 2; j++){
				if (i <= 1){
					if (j <= 1){
						Pxy[i - 1][j - 1] = CalcPnn(rho, Cx[0], Cy[0]);
					}
					else if (j >= Ky){
						Pxy[i - 1][j - 1] = CalcPnp(rho, Cx[0], Cy[Ky - 2]);
					}
					else{
						Pxy[i - 1][j - 1] = CalcPnc(rho, Cx[0], Cy[j - 2], Cy[j - 1]);
					}
				}
				else if (i >= Kx){
					if (j <= 1){
						Pxy[i - 1][j - 1] = CalcPpn(rho, Cx[Kx - 2], Cy[0]);
					}
					else if (j >= Ky){
						Pxy[i - 1][j - 1] = CalcPpp(rho, Cx[Kx - 2], Cy[Ky - 2]);
					}
					else{
						Pxy[i - 1][j - 1] = CalcPpc(rho, Cx[Kx - 2], Cy[j - 2], Cy[j - 1]);
					}
				}
				else{
					if (j <= 1){
						Pxy[i - 1][j - 1] = CalcPcn(rho, Cx[i - 2], Cx[i - 1], Cy[0]);
					}
					else if (j >= Ky){
						Pxy[i - 1][j - 1] = CalcPcp(rho, Cx[i - 2], Cx[i - 1], Cy[Ky - 2]);
					}
					else{
						Pxy[i - 1][j - 1] = CalcPcc(rho, Cx[i - 2], Cx[i - 1], Cy[j - 2], Cy[j - 1]);
					}
				}
			}
	}


}
