#include "mcmc.h"
#include "DispMessage.h"
#include "common.h" 

using namespace EstPolychoric;
using namespace cmmn;

namespace MyMCMC {

	inline double sqr(double x){
		return x * x;
	}

	double MyLog(double x){
		double v;
		if (x > 1.0e-308){
			v = Math::Log(x);
		}
		else{
			v = Math::Log(1.0e-308);
		}
		return v;
	}

	double MyExp(double x){
		double v;
		if (x < 709.0){
			v = Math::Exp(x);
		}
		else{
			v = Math::Exp(709.0);
		}
		return v;
	}


	void MCMCPolyChoric::ClearMe(){
		if(rn != nullptr) delete rn;
		if (Nkh != nullptr){
			for (int k = 0; k < Kx; k++)
				delete[] Nkh[k];
			delete[] Nkh;
		}
		if (X != nullptr) delete[] X;
		if (Y != nullptr) delete[] Y;
		if (rho != nullptr) delete[] rho;
		if (Cx != nullptr){
			delete[] Cx;
		}
		if (Cy != nullptr){
			delete[] Cy;
		}
	}

	void MCMCPolyChoric::Init(String ^ vhstr, int vnjump, int vNSimu, int vKx, int vKy, int vN, int * vX, int * vY,
		double init_rho, double * vCx, double * vCy,
		double vsgm_rho){
		hstr = vhstr;
		rn = new RNP2to191();
		if (vnjump > 0)
			for (int i = 0; i < vnjump; i++)
				rn->Jump2toj(100);
		NSimu = vNSimu;
		Kx = vKx;
		Ky = vKy;
		N = vN;
		Nkh = new int*[Kx];
		for (int k = 0; k < Kx; k++){
			Nkh[k] = new int[Ky];
			for (int h = 0; h < Ky; h++)
				Nkh[k][h] = 0;
		}

		X = new int[N];
		Y = new int[N];
		for (int i = 0; i < N; i++){
			X[i] = vX[i];
			Y[i] = vY[i];
			Nkh[X[i] - 1][Y[i] - 1]++;
		}

		rho = new double[NSimu + 1];
		rho[0] = init_rho;
		Cx = new double[Kx - 1];
		Cy = new double[Ky - 1];
		for (int k = 0; k < Kx - 1; k++)
			Cx[k] = vCx[k];
		for (int h = 0; h < Ky - 1; h++)
			Cy[h] = vCy[h];
		cnt_rho = 0;
		sgm_rho = vsgm_rho;
	}

	double MCMCPolyChoric::LProbAll(double vrho){
		CalcPxy Pxy(vrho, Kx, Ky, Cx, Cy);
		double v = 0.0;
		for (int k = 0; k < Kx; k++)
			for (int h = 0; h < Ky; h++)
				v += Nkh[k][h] * MyLog(Pxy.Pxy[k][h]);
		return v;
	}

	void MCMCPolyChoric::do_mcmc(){
		DispMessage dmfm;
		dmfm.Show();
		dmfm.SetMsg(hstr);
		dmfm.ShowMessage("do_mcmc started.");

		for (int istep = 0; istep < NSimu; istep++){
			if ((istep % 200) == 0){
				dmfm.ShowMessage("\n" + istep.ToString() + "/" + NSimu.ToString());
				Application::DoEvents();
			}

			double y = rn->normalMS(rho[istep], sgm_rho);
			double a;
			if ((y <= -1.0) || (y >= 1.0)){
				a = -1.0;
			}
			else{
				double num = LProbAll(y);
				double den = LProbAll(rho[istep]);
				a = MyExp(num - den);
			}
			if (rn->uni() < a){
				rho[istep + 1] = y;
				cnt_rho++;
			}
			else{
				rho[istep + 1] = rho[istep];
			}
		}

		dmfm.Close();
	}

}
